#include "graph.h"

// Max size of random graph name.
#define RANDOM_NAME_SIZE 64
// Initial size of lookup graph. It will double each time capacity is reached.
#define LOOKUP_GR_INIT_SIZE 64

/**
 * Extern inline functions.
 */
size_t LSUP_graph_size(LSUP_Graph *gr);

size_t LSUP_graph_capacity(LSUP_Graph *gr);


int
LSUP_graph_init(
        LSUP_Graph *gr, size_t capacity, char *uri_str,
        LSUP_store_type store_type)
{
    if (uri_str == NULL) {
        char gr_name[RANDOM_NAME_SIZE];
        sprintf(gr_name, "urn:tmp:%d", rand());
        gr->uri = LSUP_term_new(LSUP_TERM_URI, gr_name, NULL, NULL);
    } else {
        gr->uri = LSUP_term_new(LSUP_TERM_URI, uri_str, NULL, NULL);
    }

    gr->keys = LSUP_keyset_new(capacity);

    switch (store_type ) {
        case LSUP_STORE_MEM:
            gr->idx = LSUP_index_new(gr->keys->capacity);
            break;

        case LSUP_STORE_MDB:
            // TODO

        default:
            return -1;
    }

    return 0;
}


LSUP_Graph *
LSUP_graph_new(size_t capacity, char *uri_str, LSUP_store_type store_type)
{
    LSUP_Graph *gr;
    CRITICAL(gr = malloc(sizeof(LSUP_Graph)));

    LSUP_graph_init(gr, capacity, uri_str, store_type);

    return gr;
}


int
LSUP_graph_add(LSUP_Graph *gr, LSUP_Triple data[], size_t data_size)
{
    // TODO Decouple this and build interface for memory and MDB integration.

    // Resize all at once if needed.
    if (gr->keys->capacity < gr->keys->free_i + data_size)
        LSUP_keyset_resize(gr->keys, gr->keys->free_i + data_size);

    for (size_t i = 0; i < data_size; i++) {
        LSUP_SerTerm sspo[3];

        LSUP_term_serialize(data[i].s, sspo);
        LSUP_term_serialize(data[i].p, sspo + 1);
        LSUP_term_serialize(data[i].o, sspo + 2);

        LSUP_Key sk = LSUP_sterm_to_key(sspo);
        LSUP_Key pk = LSUP_sterm_to_key(sspo + 1);
        LSUP_Key ok = LSUP_sterm_to_key(sspo + 2);

        // Add terms to index.
        // TODO Optimize. This copies sterm data twice: once to serialize,
        // once to add to index.
        LSUP_index_add_pair(gr->idx, sk, sspo);
        LSUP_index_add_pair(gr->idx, pk, sspo + 1);
        LSUP_index_add_pair(gr->idx, ok, sspo + 2);

        LSUP_buffer_done(sspo);
        LSUP_buffer_done(sspo + 1);
        LSUP_buffer_done(sspo + 2);

        // Add triple.
        LSUP_TripleKey trp = {sk, pk, ok};
        LSUP_keyset_add(gr->keys, &trp, LSUP_KS_CHECK_DUP);
    }

    return 0;
}


bool
LSUP_graph_contains(const LSUP_Graph *gr, const LSUP_Triple *spo)
{
    LSUP_Key sk = LSUP_term_to_key(spo->s);
    LSUP_Key pk = LSUP_term_to_key(spo->p);
    LSUP_Key ok = LSUP_term_to_key(spo->o);

    LSUP_TripleKey spok = {sk, pk, ok};

    return LSUP_keyset_contains(gr->keys, &spok);
}


int LSUP_graph_join(LSUP_Graph *gr1, LSUP_Graph *gr2, LSUP_Graph *res)
{
    LSUP_keyset_sparse_copy(ks1, res);

    if (LSUP_keyset_seek(ks2, 0)) {
        do {
            const LSUP_TripleKey *spok = LSUP_keyset_peek(ks2);
            if (!is_null_trp(spok)) {
                LSUP_keyset_add(
                        res, spok, LSUP_KS_CHECK_DUP | LSUP_KS_CHECK_CAP);
            }
        } while (LSUP_keyset_next(ks2));
    }

    return 0;
}


int LSUP_graph_subtract(LSUP_Graph *gr1, LSUP_Graph *gr2, LSUP_Graph *res)
{
    LSUP_graph_init(res, gr1->keys->capacity, NULL, LSUP_STORE_MEM);

    if (LSUP_keyset_seek(ks1, 0)) {
        do {
            const LSUP_TripleKey *spok = LSUP_keyset_peek(ks1);
            if (!is_null_trp(spok) && !LSUP_keyset_contains(ks2, spok)) {
                LSUP_keyset_add(res, spok, 0);
            }
        } while (LSUP_keyset_next(ks1));
    }

    return 0;
}


int LSUP_graph_intersect(LSUP_Graph *gr1, LSUP_Graph *gr2, LSUP_Graph *res)
{
    LSUP_graph_init(res, gr1->keys->capacity, NULL, LSUP_STORE_MEM);

    if (LSUP_keyset_seek(ks1, 0)) {
        do {
            const LSUP_TripleKey *spok = LSUP_keyset_peek(ks1);
            if (!is_null_trp(spok) && LSUP_keyset_contains(ks2, spok)) {
                LSUP_keyset_add(res, spok, 0);
            }
        } while (LSUP_keyset_next(ks1));
    }

    return 0;
}


int LSUP_graph_xor(LSUP_Graph *gr1, LSUP_Graph *gr2, LSUP_Graph *res)
{
    LSUP_graph_init(res, gr1->keys->capacity, NULL, LSUP_STORE_MEM);

    if (LSUP_keyset_seek(ks1, 0)) {
        do {
            const LSUP_TripleKey *spok = LSUP_keyset_peek(ks1);
            if (!is_null_trp(spok) && !LSUP_keyset_contains(ks2, spok)) {
                LSUP_keyset_add(res, spok, 0);
            }
        } while (LSUP_keyset_next(ks1));
    }

    if (LSUP_keyset_seek(ks2, 0)) {
        do {
            const LSUP_TripleKey *spok = LSUP_keyset_peek(ks2);
            if (!is_null_trp(spok) && !LSUP_keyset_contains(ks1, spok)) {
                LSUP_keyset_add(res, spok, 0);
            }
        } while (LSUP_keyset_next(ks2));
    }

    return 0;
}


void
LSUP_graph_free(LSUP_Graph *gr)
{
    if(LIKELY(gr != NULL)) {
        LSUP_term_free(gr->uri);
        LSUP_keyset_free(gr->keys);
        LSUP_index_free(gr->idx);
        free(gr);
    }
}

